
static ERROR RegisterProgram(struct Install *, struct XMLTag *, OBJECTPTR, STRING);
static ERROR RegisterService(struct Install *, struct XMLTag *, OBJECTPTR, STRING);

/*****************************************************************************
** Internal: TranslateMethod()
*/

static LONG TranslateMethod(STRING String)
{
   if (StrCompare("overwrite", String, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      return(METHOD_OVERWRITE);
   }
   else if (StrCompare("backup", String, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      return(METHOD_BACKUP);
   }
   else if (StrCompare("cautious", String, 0, STR_MATCHLENGTH) IS ERR_Okay) {
      return(METHOD_CAUTIOUS);
   }
   else return(METHOD_OVERWRITE);
}

/*****************************************************************************
** Internal: CommandComponent()
*/

static ERROR CommandComponent(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   STRING name;
   LONG i;
   WORD reverse;
   ERROR error;

   DPrintF("~Installer:","Processing <component> tag.");

   name = NULL;
   reverse = FALSE;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         name = XMLTag->Attributes[i].Value;
         break;
      }
      else if (StrCompare("!name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         name = XMLTag->Attributes[i].Value;
         reverse = TRUE;
         break;
      }
   }

   if (name) {
      for (i=0; i < Self->prvTotalVariables; i++) {
         if (StrCompare(name, Self->prvVariables[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            error = StrCompare("1", Self->prvVariables[i].Value, 0, STR_MATCHLENGTH);

            if (((error IS ERR_Okay) AND (reverse IS FALSE)) OR ((error != ERR_Okay) AND (reverse IS TRUE))) {
               /*** Execute the commands within the <component> tag ***/

               DPrintF("Installer:","Criteria passed for component selection (%s = %s).", name, Self->prvVariables[i].Value);

               error = ExecuteInstallTags(Self, Output, XMLTag->Child);
            }
            else error = ERR_Okay;

            StepBack();
            return(error);
         }
      }

      DPrintF("Installer:","Criteria failed for component selection \"%s\".", name);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandCopy()
*/

static UBYTE glForceCopy;

static ERROR CommandCopy(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[260];
   struct mtDecompressFile decompress;
   struct acWrite write;
   OBJECTPTR file;
   STRING src, dest;
   LONG i, method, len, type, cflags;
   UBYTE deletedir;
   ERROR error;
   BYTE nofail, filepermissions;

   DPrintF("~Installer:","Processing <copy> tag.");

   glForceCopy = FALSE;
   src    = NULL;
   dest   = NULL;
   method = METHOD_OVERWRITE;
   nofail = FALSE;
   filepermissions = FALSE;
   Self->NoSUIDOption  = FALSE;
   Self->Permissions   = NULL;
   Self->Permanent     = FALSE;
   Self->Quiet         = 0;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("dest", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         dest = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("force", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         glForceCopy = TRUE;
      }
      else if (StrCompare("method", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         method = TranslateMethod(XMLTag->Attributes[i].Value);
      }
      else if (StrCompare("nofail", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         nofail = TRUE;
      }
      else if (StrCompare("executable", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->Permissions |= PERMIT_EVERYONE_EXEC;
      }
      else if (StrCompare("nosuid", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->NoSUIDOption = TRUE;
      }
      else if (StrCompare("permanent", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->Permanent = TRUE;
      }
      else if (StrCompare("applypermissions", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         filepermissions = TRUE;
      }
      else if (StrCompare("quiet", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->Quiet = 1; /* Do not inform the user of decompressed directory content */
      }
   }

   error = ERR_Okay;
   if ((src) AND (dest) AND (*src) AND (*dest)) {
      /* Determine if the source is from the compressed installation file,
      ** or if it is from a specific location on the user's system.
      */

      for (i=0; (src[i]) AND (src[i] != ':'); i++);

      if (src[i] != ':') {
         /* If the copy operation is to a directory and if that directory
         ** does not exist, we mark the directory for deletion during uninstall.
         */

         deletedir = FALSE;
         if (Self->prvUninstall) {
            for (i=0; dest[i]; i++);
            if ((dest[i-1] IS '/') AND (AnalyseLocation(dest, &type) != ERR_Okay)) deletedir = TRUE;
         }

         if (Self->Test IS FALSE) {
            /* 'Copy' the file by decompressing it.  Note that the source file
            ** reference may use wildcards, as supported by the DecompressFile()
            ** method.
            */

            if (filepermissions) {
               GetField(Self->prvCompression, FID_Flags, FT_LONG, &cflags);
               cflags |= CMF_APPLYSECURITY;
               SetField(Self->prvCompression, FID_Flags, FT_LONG, &cflags);
            }

            SetField(Self->prvCompression, FID_Feedback, FT_POINTER, &CopyFeedback);

               decompress.Path        = src;
               decompress.Destination = dest;
               decompress.Flags       = NULL;
               error = Action(MT_DecompressFile, Self->prvCompression, &decompress);
               if (error) DPrintF("@Install:","<copy> Error %d decompressing %s", error, src);
               if (error IS ERR_Search) error = ERR_Okay; /* Ignore wildcard problems if the requested file does not exist */

            SetField(Self->prvCompression, FID_Feedback, FT_POINTER, NULL);

            if (filepermissions) {
               cflags &= ~CMF_APPLYSECURITY;
               SetField(Self->prvCompression, FID_Flags, FT_LONG, &cflags);
            }

            if (deletedir IS TRUE) {
               write.Buffer = buffer;
               write.Length = StrFormat(buffer, sizeof(buffer), "  <delete src=\"%s\" empty/>\n", dest);
               Action(AC_Write, Self->prvUninstall, &write);
            }
         }
      }
      else {
         /*** Copy a file somewhere in the system to a new location ***/

         len = StrFormat(buffer, sizeof(buffer), "Copying \"%s\" to \"%s\".", src, dest);
         acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

         if (Self->Test IS FALSE) {
            if ((error = CreateObject(ID_FILE, NF_CHILD, &file, NULL,
                  FID_Location|TSTRING,  src,
                  FID_Flags|TLONG,       FL_READ,
                  TAGEND)) IS ERR_Okay) {
               if ((error = ActionTags(MT_CopyFile, file, dest))) {
                  DPrintF("@Install:","<copy> Error %d copying file from %s to %s", error, src, dest);
               }
               acFree(file);
            }
         }
      }
   }

   ScrollView(Self);

   StepBack();
   if (nofail) return(ERR_Okay);
   else return(error);
}

/*****************************************************************************
** Internal: CopyFeedback()
** Short:    This routine is called during the decompression of files.
*/

static LONG CopyFeedback(OBJECTPTR Compression, struct CompressionFeedback *Feedback)
{
   UBYTE buffer[160];
   struct FileTime *date;
   struct acWrite write;
   struct Install *Self;
   OBJECTPTR file;
   LARGE srcstamp, deststamp;
   LONG len, result;

   ProcessMessages(NULL, NULL);

   result = CFR_Okay;
   if (Feedback->Progress IS 0) {
      if (AccessObject(GetOwner(Compression), 0, (OBJECTPTR *)&Self) IS ERR_Okay) {

         if (Self->CancelRequest) {
            DPrintF("@Install:","Cancellation requested, aborting install during decompression.");
            ReleaseObject(Self);
            return(CFR_Abort);
         }

         if (glForceCopy IS FALSE) {
            /* Check the file date stamps of the source and destination.  If the
            ** destination is more recent than the source, do not overwrite it.
            */

            if (AnalyseLocation(Feedback->Destination, NULL) IS ERR_Okay) {
               if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
                  FID_Location|TSTRING, Feedback->Destination,
                  TAGEND) IS ERR_Okay) {

                  if ((GetField(file, FID_Date, FT_POINTER, &date) IS ERR_Okay) AND (date)) {
                     /*** Is the destination more recent than the source? ***/

                     //DPrintF("Install:","Src: %d/%d/%d %d:%d:%d", Feedback->Day, Feedback->Month, Feedback->Year, Feedback->Hour, Feedback->Minute, Feedback->Second);
                     //DPrintF("Install:","Dst: %d/%d/%d %d:%d:%d", date->Day, date->Month, date->Year, date->Hour, date->Minute, date->Second);

                     srcstamp = ((LARGE)Feedback->Minute * 60) +
                              ((LARGE)Feedback->Hour * 60 * 60) +
                              ((LARGE)Feedback->Day * 60 * 60 * 24) +
                              ((LARGE)Feedback->Month * 60 * 60 * 24 * 31) +
                              ((LARGE)Feedback->Year * 60 * 60 * 24 * 31 * 12);

                     deststamp = ((LARGE)date->Minute * 60) +
                              ((LARGE)date->Hour * 60 * 60) +
                              ((LARGE)date->Day * 60 * 60 * 24) +
                              ((LARGE)date->Month * 60 * 60 * 24 * 31) +
                              ((LARGE)date->Year * 60 * 60 * 24 * 31 * 12);

                     if (deststamp > srcstamp) result = CFR_Skip;
                  }

                  acFree(file);
               }
            }
         }

         /*** Send output to the user ***/

         if (result IS CFR_Skip) {
            if (Self->Quiet < 2) {
               len = StrFormat(buffer, sizeof(buffer), "Skipping file \"%s\" - destination more recent.", Feedback->Path, Feedback->Destination);
               acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);
            }
         }
         else {
            if (Self->Quiet < 2) {
               if (Self->Permissions & PERMIT_EVERYONE_EXEC) {
                  len = StrFormat(buffer, sizeof(buffer), "Installing \"%s\" [exe]", Feedback->Destination);
               }
               else len = StrFormat(buffer, sizeof(buffer), "Installing \"%s\"", Feedback->Destination);
               acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);
            }

            /*** Add the equivalent uninstall command ***/

            if ((Self->prvUninstall) AND (Self->Permanent IS FALSE)) {
               for (len=0; Feedback->Destination[len]; len++);
               if ((Feedback->Destination[len-1] IS '/') OR (Feedback->Destination[len-1] IS '\\')) {
                  write.Length = StrFormat(buffer, sizeof(buffer), "  <delete src=\"%s\" empty/>\n", Feedback->Destination);
               }
               else write.Length = StrFormat(buffer, sizeof(buffer), "  <delete src=\"%s\"/>\n", Feedback->Destination);
               write.Buffer = buffer;
               Action(AC_Write, Self->prvUninstall, &write);
            }
         }

         ReleaseObject(Self);
      }
   }
   else if (Feedback->Progress IS 100) {
      if (AccessObject(GetOwner(Compression), 0, (OBJECTPTR *)&Self) IS ERR_Okay) {
         if (Self->Permissions & PERMIT_EVERYONE_EXEC) {
#ifdef _WIN32
            /*** Exe permissions are not used in Win32 ***/
#else
            DPrintF("Install:","Setting execute permissions on \"%s\"", Feedback->Destination);

            /* Open the existing file, read its permissions and then
            ** set the execution bit.
            */

            if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
               FID_Location|TSTRING, Feedback->Destination,
               TAGEND) IS ERR_Okay) {
               LONG permissions;

               if (GetField(file, FID_Permissions, FT_LONG, &permissions) IS ERR_Okay) {
                  /* Note that the USERID bit is set so that the installed file will
                  ** run with the same permissions as the logged in user.  This means
                  ** that if the user logs in as root and installs program files, they
                  ** will be able to run with fast video access (due to 'suid root'
                  ** security access to /dev/mem).
                  */
                  permissions |= PERMIT_EVERYONE_EXEC;

                  #ifdef __linux__
                     if (Self->NoSUIDOption IS FALSE) {
                        if (AnalyseLocation("/dev/pcmem", NULL) != ERR_Okay) permissions |= PERMIT_USERID;
                     }
                  #endif

                  if (SetField(file, FID_Permissions, FT_LONG, &permissions) != ERR_Okay) {
                     UBYTE str[] = "<item><font colour=\"#a02020\">Failed to set EXE permissions on the destination file.</font></item>";
                     acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_XML, 1, str, sizeof(str), 1);
                  }
               }
               else {
                  UBYTE str[] = "<item><font colour=\"#a02020\">Failed to read file permissions on the destination file.</font></item>";
                  acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_XML, 1, str, sizeof(str), 1);
               }
               acFree(file);
            }
            else {
               UBYTE str[] = "<item><font colour=\"#a02020\">Failed to open the file for setting the EXE bit.</font></item>";
               acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_XML, 1, str, sizeof(str), 1);
            }
#endif
         }
         ReleaseObject(Self);
      }
   }

   ScrollView(Self);

   if (Self->Quiet IS 1) Self->Quiet++;

   return(result);
}

/*****************************************************************************
** Internal: CommandDelete()
**
** <delete src="athene:reamde.txt"/>
*/

static ERROR CommandDelete(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   WORD i, len;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         len = StrFormat(buffer, sizeof(buffer), "Deleting \"%s\".", XMLTag->Attributes[i].Value);
         ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

         if (!Self->Test) DeleteFile(XMLTag->Attributes[i].Value);
      }
   }

   ScrollView(Self);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandDirectory()
*/

static ERROR CommandDirectory(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct XMLTag *tag;
   OBJECTPTR info, input;
   OBJECTID browse_id, dialog_id, onclick_id;
   STRING content;
   LONG i, height;

   Self->prvDirectories[Self->prvDirectoryCount].Name = NULL;
   Self->prvDirectories[Self->prvDirectoryCount].Path = NULL;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("path", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvDirectories[Self->prvDirectoryCount].Path = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvDirectories[Self->prvDirectoryCount].Name = XMLTag->Attributes[i].Value;
      }
      else DPrintF("@Installer:","[Install] Unsupported directory attribute: %s.", XMLTag->Attributes[i].Name);
   }

   content = NULL;
   for (tag=XMLTag->Child; tag; tag=tag->Next) {
      if (!tag->Attributes[0].Name) { content = tag->Attributes[0].Value; break; }
   }

   if ((Self->prvDirectories[Self->prvDirectoryCount].Name) AND (Self->prvDirectories[Self->prvDirectoryCount].Path)) {

      if (content) {
         CreateObject(ID_TEXT, NULL, &info, NULL,
            FID_Owner|TLONG,     Self->ViewID,
            FID_String|TSTRING,  content,
            FID_XCoord|TLONG,    Self->prvCurrentXCoord,
            FID_YCoord|TLONG,    Self->prvCurrentYCoord,
            FID_Face|TSTRING,    glFontFace,
            FID_Colour|TSTRING,  "#000000",
            TAGEND);

         GetField(info, FID_TextHeight, FT_LONG, &height);
         Self->prvCurrentYCoord += height + 10;

         ActionMsg(AC_Draw, Self->ViewID, NULL);
      }

      CreateObject(ID_INPUT, NULL, &input, NULL,
         FID_Owner|TLONG,       Self->ViewID,
         FID_XCoord|TLONG,      Self->prvCurrentXCoord + 20,
         FID_XOffset|TLONG,     20,
         FID_YCoord|TLONG,      Self->prvCurrentYCoord,
         FID_Height|TLONG,      20,
         FID_String|TSTRING,    Self->prvDirectories[Self->prvDirectoryCount].Path,
         FID_Highlight|TSTRING, "#202020",
         FID_Shadow|TSTRING,    "#202020",
         FID_Colour|TSTRING,    "#f0f0f0",
         TAGEND);

      Self->prvDirectories[Self->prvDirectoryCount].Text = input;

      /*** Recalculate the render object height based on the height of our font ***/

      GetField(input, FID_Height, FT_LONG, &height);

      acShow(input);

      /*** Create a directory browse button ***/

      CreateObject(ID_RENDER, NULL, NULL, &browse_id,
         FID_Owner|TLONG,   Self->ViewID,
         FID_XOffset|TLONG, 0,
         FID_YCoord|TLONG,  Self->prvCurrentYCoord,
         FID_Height|TLONG,  height,
         FID_Width|TLONG,   16,
         TAGEND);

      CreateObject(ID_IMAGE, NULL, NULL, NULL,
         FID_Owner|TLONG,   browse_id,
         FID_Src|TSTRING,   "icons:folders/folder(16)",
         FID_Align|TSTRING, "vertical",
         TAGEND);

      CreateObject(ID_ONCLICK, NULL, NULL, &onclick_id,
         FID_Owner|TLONG, browse_id,
         TAGEND);

      CreateObject(ID_FILEDIALOG, NULL, NULL, &dialog_id,
         FID_Owner|TLONG,      onclick_id,
         FID_Static|TLONG,     TRUE,
         FID_Title|TSTRING,    "Select a Destination Folder",
         FID_Location|TSTRING, Self->prvDirectories[Self->prvDirectoryCount].Path,
         FID_Flags|TLONG,      FDF_SELECTDIR|FDF_NOFILES,
         FID_OkayText|TSTRING, "Select Dir",
         TAGEND);

      CreateObject(ID_SET, NULL, NULL, NULL,
         FID_Owner|TLONG,      dialog_id,
         FID_Object|TLONG,     input->UniqueID,
         FID_String|TUNLISTED, "[owner.location]",
         TAGEND);

      acShowID(browse_id);

      Self->prvCurrentYCoord += height + 14;

      Self->prvDirectoryCount++;
   }
   else DPrintF("@Installer:","[Install] Error in <destination> tag, missing compulsory attributes.");

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandItem()
*/

static ERROR CommandItem(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[200];
   struct acWrite write;
   struct XMLTag *tag;
   STRING content, item;
   LONG i;
   WORD method;

   if (!Self->prvConfig) return(ERR_Okay);
   if (Self->Test) return(ERR_Okay);

   item   = NULL;
   method = METHOD_OVERWRITE;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         item = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("method", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         method = TranslateMethod(XMLTag->Attributes[i].Value);
      }
      else DPrintF("@Installer:","[Install] Unsupported <item> attribute: %s.", XMLTag->Attributes[i].Name);
   }

   content = NULL;
   for (tag=XMLTag->Child; tag; tag=tag->Next) {
      if (!tag->Attributes[0].Name) { content = tag->Attributes[0].Value; break; }
   }

   if (Self->prvUninstall) {
      /* If the item already exists then record its value prior to us overwriting it.
      ** Otherwise, just tell the uninstaller to delete the item.
      */

      write.Buffer = buffer;
      write.Length = StrFormat(buffer, sizeof(buffer), "    <item name=\"%s\"/>\n", item);
      if (Action(AC_Write, Self->prvUninstall, &write) != ERR_Okay) return(ERR_Write);
   }

   ActionTags(MT_WriteConfig, Self->prvConfig, Self->prvMergeSection, item, content);

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandLink()
*/

static ERROR CommandLink(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct acWrite write;
   UBYTE buffer[500];
   LONG i, len;
   STRING from, to;
   ERROR error;
   BYTE permanent;

   from = NULL;
   to = NULL;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("from", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         from = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("to", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         to = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("permanent", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         permanent = TRUE;
      }
   }

   if ((from) AND (to)) {
      len = StrFormat(buffer, sizeof(buffer), "Linking \"%s\" to \"%s\".", from, to);
      acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      if (Self->Test IS FALSE) {
         DeleteFile(from); /* Remove any existing link first */
         error = CreateLink(from, to);
      }
      else error = ERR_Okay;

      if (error IS ERR_Okay) {
         if ((Self->prvUninstall) AND (Self->Permanent IS FALSE)) {
            write.Length = StrFormat(buffer, sizeof(buffer), "  <delete src=\"%s\"/>\n", from);
            write.Buffer = buffer;
            Action(AC_Write, Self->prvUninstall, &write);
         }
         return(ERR_Okay);
      }
      else {
         len = StrFormat(buffer, sizeof(buffer), "<item><font colour=\"#a02020\">Failed to create link at \"%s\",  error %d.</font></item>", from, error);
         acDataChannel(Output, Self->Head.UniqueID, DATA_XML, 1, buffer, len + 1, 1);
         return(error);
      }
   }
   else {
      len = StrCopy("<item><font colour=\"#a02020\">Link tag is missing From and/or To attributes.</font></item>", buffer, sizeof(buffer));
      acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);
      return(ERR_Failed);
   }
}

/*****************************************************************************
** Internal: CommandMerge()
*/

static ERROR CommandMerge(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[300];
   struct acWrite write;
   STRING file;
   LONG i, method, nofail, len;
   ERROR error;

   if (!XMLTag->Attributes) return(ERR_Okay);

   DPrintF("~Installer:","Processing <merge> tag.");

   file    = NULL;
   Self->prvMergeSection = NULL;
   method  = METHOD_OVERWRITE;
   nofail  = FALSE;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("file", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         file = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("section", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvMergeSection = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("method", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         method = TranslateMethod(XMLTag->Attributes[i].Value);
      }
      else if (StrCompare("nofail", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         nofail = TRUE;
         Self->NoFail++;
      }
   }

   error = ERR_Okay;
   if ((file) AND (Self->prvMergeSection)) {
      /*** Tell the user what we are doing ***/

      len = StrFormat(buffer, sizeof(buffer), "Merging section \"%s\" into configuration file \"%s\".", Self->prvMergeSection, file);
      acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      if (Self->Test IS TRUE) goto exit;

      if (Self->prvUninstall) {
         write.Length = StrFormat(buffer, sizeof(buffer), "  <merge file=\"%s\" section=\"%s\">\n", file, Self->prvMergeSection);
         write.Buffer = buffer;
         if ((error = Action(AC_Write, Self->prvUninstall, &write)) != ERR_Okay) goto exit;
      }

      /*** Open the configuration file ***/

      if ((error = CreateObject(ID_CONFIG, NF_CHILD, &Self->prvConfig, NULL,
            FID_Location|TSTRING, file,
            TAGEND)) IS ERR_Okay) {

         if ((error = ExecuteInstallTags(Self, Output, XMLTag->Child)) != ERR_Okay) goto exit;

         /*** Save configuration file changes ***/
         SaveObjectToFile(Self->prvConfig, file, PERMIT_EVERYONE_READWRITE);

         acFree(Self->prvConfig);
         Self->prvConfig = NULL;
      }

      if (Self->prvUninstall) {
         write.Buffer = "  </merge>\n";
         for (i=0; write.Buffer[i]; i++);
         write.Length = i;
         if ((error = Action(AC_Write, Self->prvUninstall, &write)) != ERR_Okay) goto exit;
      }
   }
   else DPrintF("@Installer:","[Install] A <method> tag is missing file and/or section attributes.");

exit:
   StepBack();
   if (nofail) {
      Self->NoFail--;
      return(ERR_Okay);
   }
   else return(error);
}

/*****************************************************************************
** Internal: CommandMove()
*/

static ERROR CommandMove(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[300];
   LONG i, len;
   STRING src, dest;
   ERROR error;

   src = NULL;
   dest = NULL;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("dest", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         dest = XMLTag->Attributes[i].Value;
      }
   }

   if ((src) AND (dest)) {
      len = StrFormat(buffer, sizeof(buffer), "Moving \"%s\" to \"%s\".", src, dest);
      acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      error = MoveLocation(src, dest);

      if (error IS ERR_Okay) return(ERR_Okay);
      else {
         len = StrFormat(buffer, sizeof(buffer), "<item><font colour=\"#a02020\">Failed to move file \"%s\", error %d.</font></item>", src, error);
         acDataChannel(Output, Self->Head.UniqueID, DATA_XML, 1, buffer, len + 1, 1);
         DPrintF("@Install:", buffer);
         return(error);
      }
   }
   else {
      len = StrCopy("<item><font colour=\"#a02020\">Move tag is missing src and/or dest attributes.</font></item>", buffer, sizeof(buffer));
      acDataChannel(Output, Self->Head.UniqueID, DATA_XML, 1, buffer, len + 1, 1);
      DPrintF("@Install:", buffer);
      return(ERR_Failed);
   }
}

/*****************************************************************************
** Internal: CommandNoFail()
*/

static ERROR CommandNoFail(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   DPrintF("~Installer:","Processing <nofail> tag.");

   Self->NoFail++;
   ExecuteInstallTags(Self, Output, XMLTag->Child);
   Self->NoFail--;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandOption()
*/

static ERROR CommandOption(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   STRING text;
   LONG i;
   ERROR error;

   Self->prvOptions[Self->prvOptionCount].Name = NULL;
   Self->prvOptions[Self->prvOptionCount].OptionStatus = FALSE;
   text = NULL;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("value", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (StrCompare("on", XMLTag->Attributes[i].Value, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            Self->prvOptions[Self->prvOptionCount].OptionStatus = TRUE;
         }
         else Self->prvOptions[Self->prvOptionCount].OptionStatus = FALSE;
      }
      else if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvOptions[Self->prvOptionCount].Name = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("text", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         text = XMLTag->Attributes[i].Value;
      }
      else DPrintF("@Installer:","[Install] Unsupported <option> attribute: %s.", XMLTag->Attributes[i].Name);
   }

   if ((Self->prvOptions[Self->prvOptionCount].Name) AND (text)) {
      Self->prvOptions[Self->prvOptionCount].Level = Self->prvIndentLevel;
      Self->prvOptions[Self->prvOptionCount].XCoord = Self->prvCurrentXCoord;
      Self->prvOptions[Self->prvOptionCount].YCoord = Self->prvCurrentYCoord;
      if (CreateOption(Self, &Self->prvOptions[Self->prvOptionCount].ButtonID, text, Self->prvOptions[Self->prvOptionCount].OptionStatus) IS ERR_Okay) {
         Self->prvOptionCount++;
      }
   }
   else DPrintF("@Installer:","[Install] Error in <option> tag, missing compulsory attributes.");

   Self->prvCurrentXCoord += Self->prvIndent;
   Self->prvIndentLevel++;

   error = ExecuteInstallTags(Self, Output, XMLTag->Child);

   Self->prvIndentLevel--;
   Self->prvCurrentXCoord -= Self->prvIndent;

   return(error);
}

/*****************************************************************************
** Internal: CommandRegister()
*/

static ERROR CommandRegister(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct acWrite write;
   UBYTE buffer[300];
   OBJECTPTR config, userconfig;
   OBJECTID systemclasses;
   STRING classname, classid, parentid, category, dest, src, program, extension, header;
   STRING filetype, open, edit, view, service;
   LONG i, len, exists;
   ERROR error;

   if (!XMLTag->Attributes) return(ERR_Okay);

   DPrintF("~Installer:","Processing <register> tag.");

   classname = NULL;
   parentid  = NULL;
   classid   = NULL;
   category  = NULL;
   extension = NULL;
   header    = NULL;
   dest      = NULL;
   src       = NULL;
   program   = NULL;
   service   = NULL;
   filetype  = NULL;
   open      = NULL;
   edit      = NULL;
   view      = NULL;
   exists    = FALSE;

   /*** Scan for tags that are valid when registering a class ***/

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("class", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         classname = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("id", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         classid = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("parentid", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         parentid = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("extension", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         extension = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("header", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         header = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("category", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         category = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("dest", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         dest = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("program", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         program = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("service", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         service = XMLTag->Attributes[i].Value;
      }
      /*** FILETYPES ***/
      else if (StrCompare("filetype", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         filetype = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("open", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         open = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("view", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         view = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("edit", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         edit = XMLTag->Attributes[i].Value;
      }
   }

   error = ERR_Okay;
   if (filetype) {
      len = StrFormat(buffer, sizeof(buffer), "Registering filetype \"%s\".", classname);
      acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      /* By default, filetypes are registered globally (this works if
      ** the user is logged in as an administrator).  If saving to the
      ** file fails, we register the filetype in the user's personal
      ** filetype assocations instead.
      **
      ** NOTE: Any failure to register a filetype is considered non-fatal.
      */

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "system:config/software/associations.cfg",
         TAGEND) IS ERR_Okay) {

         if (extension) ActionTags(MT_WriteConfig, config, filetype, "Extension", extension);
         if (classname) ActionTags(MT_WriteConfig, config, filetype, "Class", extension);
         if (open) ActionTags(MT_WriteConfig, config, filetype, "Open", extension);
         if (view) ActionTags(MT_WriteConfig, config, filetype, "View", extension);
         if (edit) ActionTags(MT_WriteConfig, config, filetype, "Edit", extension);

         error = SaveObjectToFile(config, "system:config/software/associations.cfg", NULL);

         acFree(config);
      }

      if (error != ERR_Okay) {
         if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
            FID_Location|TSTRING, "user:config/associations.cfg",
            TAGEND) IS ERR_Okay) {

            if (extension) ActionTags(MT_WriteConfig, config, filetype, "Extension", extension);
            if (classname) ActionTags(MT_WriteConfig, config, filetype, "Class", extension);
            if (open) ActionTags(MT_WriteConfig, config, filetype, "Open", extension);
            if (view) ActionTags(MT_WriteConfig, config, filetype, "View", extension);
            if (edit) ActionTags(MT_WriteConfig, config, filetype, "Edit", extension);

            error = SaveObjectToFile(config, "user:config/associations.cfg", NULL);

            acFree(config);
         }
      }

      error = ERR_Okay;
   }
   else if ((classname) AND (classid) AND (category) AND (dest) AND (src)) {
      len = StrFormat(buffer, sizeof(buffer), "Registering class \"%s\".", classname);
      acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      if ((error = FastFindObject("SystemClasses", ID_CONFIG, &systemclasses, 1, NULL)) IS ERR_Okay) {
         if ((error = AccessObject(systemclasses, 5000, &config)) IS ERR_Okay) {
            if (parentid) ActionTags(MT_WriteConfig, config, classname, "ParentID", parentid);
            ActionTags(MT_WriteConfig, config, classname, "ClassID", classid);
            ActionTags(MT_WriteConfig, config, classname, "Category", category);
            ActionTags(MT_WriteConfig, config, classname, "Location", dest);
            if (extension) ActionTags(MT_WriteConfig, config, classname, "Extension", extension);
            if (header) ActionTags(MT_WriteConfig, config, classname, "Header", header);
            ReleaseObject(config);
         }
         else goto exit;

         if ((error = CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
            FID_Location|TSTRING, "system:config/kernel/classes.cfg",
            TAGEND)) IS ERR_Okay) {

            if (parentid) ActionTags(MT_WriteConfig, config, classname, "ParentID", parentid);
            ActionTags(MT_WriteConfig, config, classname, "ClassID", classid);
            ActionTags(MT_WriteConfig, config, classname, "Category", category);
            ActionTags(MT_WriteConfig, config, classname, "Location", dest);
            if (extension) ActionTags(MT_WriteConfig, config, classname, "Extension", extension);
            if (header) ActionTags(MT_WriteConfig, config, classname, "Header", header);

            SaveObjectToFile(config, "system:config/kernel/classes.cfg", PERMIT_EVERYONE_READWRITE);

            acFree(config);
         }
         else goto exit;

         error = ActionTags(MT_DecompressFile, Self->prvCompression, src, dest, NULL);
      }
      else goto exit;

      if ((Self->prvUninstall) AND (exists IS FALSE)) {
         write.Length = StrFormat(buffer, sizeof(buffer), "  <unregister class=\"%s\" location=\"%s\"/>\n", classname, dest);
         write.Buffer = buffer;
         Action(AC_Write, Self->prvUninstall, &write);
      }
   }
   else if (program) {
      /*** Tell the user what we are doing ***/

      len = StrFormat(buffer, sizeof(buffer), "Registering program \"%s\".", program);
      ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      /*** Register the program in the global program configuration file ***/

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "system:config/software/programs.cfg",
         TAGEND) IS ERR_Okay) {

         RegisterProgram(Self, XMLTag, config, program);

         if (SaveObjectToFile(config, "system:config/software/programs.cfg", PERMIT_EVERYONE_READWRITE) != ERR_Okay) {
            /* If the user does not have write access to the global programs.cfg
            ** file, we add the configuration information to his personal
            ** programs.cfg file instead.
            */

            if (CreateObject(ID_CONFIG, NF_CHILD, &userconfig, NULL,
               FID_Location|TSTRING, "user:config/programs.cfg",
               TAGEND) IS ERR_Okay) {

               RegisterProgram(Self, XMLTag, userconfig, program);

               SaveObjectToFile(userconfig, "user:config/programs.cfg", PERMIT_READ|PERMIT_WRITE);

               acFree(userconfig);
            }
         }

         acFree(config);
      }

      if (Self->prvUninstall) {
         write.Buffer = buffer;
         write.Length = StrFormat(buffer, sizeof(buffer), "  <unregister program=\"%s\"/>\n", program);
         Action(AC_Write, Self->prvUninstall, &write);
      }
   }
   else if (service) {
      /*** Tell the user what we are doing ***/

      len = StrFormat(buffer, sizeof(buffer), "Registering service \"%s\".", service);
      acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

      /*** Register the service in the global service configuration file ***/

      if (CreateObject(ID_CONFIG, NF_CHILD, &config, NULL,
         FID_Location|TSTRING, "system:config/software/services.cfg",
         TAGEND) IS ERR_Okay) {

         RegisterService(Self, XMLTag, config, service);

         if (SaveObjectToFile(config, "system:config/software/services.cfg", PERMIT_EVERYONE_READWRITE) != ERR_Okay) {
            /* If the user does not have write access to the global services.cfg
            ** file, write an error message.
            */
            len = StrFormat(buffer, sizeof(buffer), "<item><font colour=\"#a02020\">Failed to register service \"%s\" as you are not an administrator.</font></item>", service);
            acDataChannel(Output, Self->Head.UniqueID, DATA_XML, 1, buffer, len + 1, 1);
            DPrintF("@Install:", buffer);
         }

         acFree(config);
      }

      if (Self->prvUninstall) {
         write.Buffer = buffer;
         write.Length = StrFormat(buffer, sizeof(buffer), "  <unregister service=\"%s\"/>\n", service);
         Action(AC_Write, Self->prvUninstall, &write);
      }
   }
   else DPrintF("@Installer:","[Install] A <register> tag is missing required attributes.");

exit:
   StepBack();
   return(error);
}

/****************************************************************************/

static ERROR RegisterProgram(struct Install *Self, struct XMLTag *XMLTag, OBJECTPTR Config, STRING Section)
{
   struct XMLTag *tag;
   STRING content, cmdfile, args, options, filetype, mode;
   WORD i;

   /* Mark the desktop for refresh.  This required for two situations: Either the
   ** program has an icon on the desktop, or the desktop has a popup menu that
   ** contains a list of all system programs.
   */

   Self->prvRefreshDesktop = TRUE;
   cmdfile = NULL;
   args = NULL;
   options = NULL;
   filetype = NULL;
   mode = NULL;

   for (tag=XMLTag->Child; tag; tag=tag->Next) {
      if (!tag->Child) continue;
      if (!(content = tag->Child->Attributes[0].Value)) continue;

      if (StrCompare("desktop", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if ((StrCompare("1", content, 0, STR_MATCHLENGTH) IS ERR_Okay) OR (StrCompare("Yes", content, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
            ActionTags(MT_WriteConfig, Config, Section, "Desktop", "Yes");
         }
         else ActionTags(MT_WriteConfig, Config, Section, "Desktop", "No");
      }
      else if (StrCompare("text", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Text", content);
      }
      else if (StrCompare("description", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Description", content);
      }
      else if (StrCompare("icon", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Icon", content);
      }
      else if (StrCompare("options", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Options", content);
         options = content;
      }
      else if (StrCompare("filetype", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Type", content);
         filetype = content;
      }
      else if (StrCompare("mode", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Mode", content);
         mode = content;
      }
      else if (StrCompare("command", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (StrCompare("run ", content, 4, NULL) IS ERR_Okay) {
            /* The use of "run src="..." args="..."" is deprecated, but we need
            ** to support it for backwards compatibility.
            */

            ActionTags(MT_WriteConfig, Config, Section, "Command", content);

            /*** Split the run instruction into commandfile and args parts ***/

            if ((i = StrSearch("src=\"", content, FALSE)) >= 0) {
               content += i + 5;
               for (i=0; (content[i]) AND (content[i] != '"'); i++);
               {
                  UBYTE buffer[i+1];
                  for (i=0; (content[i]) AND (content[i] != '"'); i++) buffer[i] = content[i];
                  buffer[i] = 0;
                  ActionTags(MT_WriteConfig, Config, Section, "CommandFile", buffer);
               }
            }

            if ((i = StrSearch("args=\"", content, FALSE)) >= 0) {
               content += i + 6;
               for (i=0; (content[i]) AND (content[i] != '"'); i++);
               {
                  UBYTE buffer[i+1];
                  for (i=0; (content[i]) AND (content[i] != '"'); i++) buffer[i] = content[i];
                  buffer[i] = 0;
                  ActionTags(MT_WriteConfig, Config, Section, "Args", buffer);
               }
            }

            if ((i = StrSearch("flags=\"", content, FALSE)) >= 0) {
               content += i + 7;
               for (i=0; (content[i]) AND (content[i] != '"'); i++);
               {
                  UBYTE buffer[i+1];
                  for (i=0; (content[i]) AND (content[i] != '"'); i++) buffer[i] = content[i];
                  buffer[i] = 0;
                  ActionTags(MT_WriteConfig, Config, Section, "Options", buffer);
               }
            }
         }
         else {
            ActionTags(MT_WriteConfig, Config, Section, "CommandFile", content);
            cmdfile = content;
         }
      }
      else if (StrCompare("args", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Args", content);
         args = content;
      }
      else if (StrCompare("category", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Category", content);
      }
      else DPrintF("@Installer:","[Install] Unrecognised <register> child tag \"%s\".", tag->Attributes[0].Name);
   }

   if (cmdfile) {
      UBYTE buffer[1024];
      i = StrFormat(buffer, sizeof(buffer), "run src=\"%s\"", cmdfile);
      if (args)     i += StrFormat(buffer+i, sizeof(buffer)-i, " args=\"%s\"", args);
      if (options)  i += StrFormat(buffer+i, sizeof(buffer)-i, " flags=\"!%s\"", options);
      if (mode)     i += StrFormat(buffer+i, sizeof(buffer)-i, " mode=\"%s\"", mode);
      if (filetype) i += StrFormat(buffer+i, sizeof(buffer)-i, " type=\"%s\"", filetype);
      ActionTags(MT_WriteConfig, Config, Section, "Command", buffer);
   }

   if (Self->prvPackageName) {
      ActionTags(MT_WriteConfig, Config, Section, "Package", Self->prvPackageName);
   }

   return(ERR_Okay);
}

/****************************************************************************/

static ERROR RegisterService(struct Install *Self, struct XMLTag *XMLTag, OBJECTPTR Config, STRING Section)
{
   struct XMLTag *tag;
   STRING content, cmdfile, args;

   cmdfile = NULL;
   args = NULL;

   for (tag=XMLTag->Child; tag; tag=tag->Next) {
      if (!tag->Child) continue;
      if (!(content = tag->Child->Attributes[0].Value)) continue;

      if (StrCompare("active", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if ((StrCompare("1", content, 0, STR_MATCHLENGTH) IS ERR_Okay) OR (StrCompare("Yes", content, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
            ActionTags(MT_WriteConfig, Config, Section, "Active", "Yes");
         }
         else ActionTags(MT_WriteConfig, Config, Section, "Active", "No");
      }
      else if (StrCompare("text", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Text", content);
      }
      else if (StrCompare("description", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Description", content);
      }
      else if (StrCompare("icon", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Icon", content);
      }
      else if (StrCompare("command", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Command", content);
         cmdfile = content;
      }
      else if (StrCompare("args", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Args", content);
         args = content;
      }
      else if (StrCompare("category", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         ActionTags(MT_WriteConfig, Config, Section, "Category", content);
      }
      else DPrintF("@Installer:","[Install] Unrecognised <register> child tag \"%s\".", tag->Attributes[0].Name);
   }

   if (Self->prvPackageName) {
      ActionTags(MT_WriteConfig, Config, Section, "Package", Self->prvPackageName);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: CommandRename()
*/

static ERROR CommandRename(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR file;
   LONG i, len;
   STRING src, newname;
   ERROR error;

   src = NULL;
   newname = NULL;
   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         src = XMLTag->Attributes[i].Value;
      }
      else if (StrCompare("newname", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         newname = XMLTag->Attributes[i].Value;
      }
   }

   error = ERR_Okay;
   if ((src) AND (newname)) {
      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, src,
         TAGEND) IS ERR_Okay) {

         len = StrFormat(buffer, sizeof(buffer), "Renaming \"%s\" to \"%s\".", src, newname);
         acDataChannel(Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);

         if (Self->Test IS FALSE) error = ActionTags(AC_Rename, file, newname);

         acFree(file);
         return(error);
      }
      else {
         len = StrFormat(buffer, sizeof(buffer), "<item><font colour=\"#a02020\">Failed to open file \"%s\" for renaming.</font></item>", src);
         acDataChannel(Output, Self->Head.UniqueID, DATA_XML, 1, buffer, len + 1, 1);
         DPrintF("@Install:", buffer);
         return(ERR_File);
      }
   }
   else {
      len = StrCopy("<item><font colour=\"#a02020\">Rename tag is missing src and/or newname attributes.</font></item>", buffer, sizeof(buffer));
      acDataChannel(Output, Self->Head.UniqueID, DATA_XML, 1, buffer, len + 1, 1);
      DPrintF("@Install:", buffer);
      return(ERR_Failed);
   }
}

/*****************************************************************************
** Internal: CommandRun()
*/

static ERROR CommandRun(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   UBYTE buffer[160];
   OBJECTPTR run;
   STRING src, args;
   ERROR error;
   LONG i, len, nofail;

   src = NULL;
   args = "";
   nofail = FALSE;
   if (NewObject(ID_RUN, NF_CHILD, &run, NULL) IS ERR_Okay) {

      for (i=1; i < XMLTag->TotalAttributes; i++) {
         if (StrCompare("src", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetFieldVariable(run, "src", XMLTag->Attributes[i].Value);
            src = XMLTag->Attributes[i].Value;
         }
         else if (StrCompare("args",  XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetField(run, FID_Args, FT_STRING, XMLTag->Attributes[i].Value);
            args = XMLTag->Attributes[i].Value;
         }
         else if (StrCompare("mode", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetFieldVariable(run, "mode", XMLTag->Attributes[i].Value);
         }
         else if (StrCompare("flags", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            SetFieldVariable(run, "flags", XMLTag->Attributes[i].Value);
         }
         else if (StrCompare("nofail", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            nofail = TRUE;
         }
      }

      /*** Tell the user what we are doing ***/

      if (src) {
         len = StrFormat(buffer, sizeof(buffer), "%s %s", src, args);
         ActionTags(AC_DataChannel, Output, Self->Head.UniqueID, DATA_TEXT, 1, buffer, len + 1, 1);


         if (Self->Test IS FALSE) {
            if ((error = acInit(run)) IS ERR_Okay) {
               if ((error = acActivate(run)) IS ERR_Okay) {
                  acFree(run);
                  return(ERR_Okay);
               }
            }
         }
         else error = ERR_Okay;
      }
      else error = ERR_Okay;

      acFree(run);
   }
   else error = ERR_NewObject;

   if (nofail IS TRUE) return(ERR_Okay);
   else return(error);
}

/*****************************************************************************
** Internal: CommandSection()
*/

static ERROR CommandSection(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct acWrite write;
   UBYTE buffer[80];
   LONG i;

   for (i=1; i < XMLTag->TotalAttributes; i++) {
      if (StrCompare("name", XMLTag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         Self->prvMergeSection = XMLTag->Attributes[i].Value;
      }
      else DPrintF("@Installer:","[Install] Unsupported <section> attribute: %s.", XMLTag->Attributes[i].Name);
   }

   if (Self->prvUninstall) {
      write.Length = StrFormat(buffer, sizeof(buffer), "  <section name=\"%s\"/>\n", Self->prvMergeSection);
      write.Buffer = buffer;
      return(Action(AC_Write, Self->prvUninstall, &write));
   }

   return(ERR_Okay);
}
